__all__ = ["Observer"]

import time as pytime
from functools import lru_cache
from typing import Optional, Sequence, Union

import astropy.units as u
from astropy.coordinates import LSR, CartesianDifferential, EarthLocation, SkyCoord
from astropy.time import Time

from ..typing import CoordFrameType, QuantityValue, UnitType


@lru_cache
def get_v_bary():
    """Barycentric velocity of the sun relative to the LSR; towards the solar apex."""
    SolarApex = SkyCoord(
        ra="18h",
        dec="30d",
        frame="fk4",
        equinox=Time("B1900"),
    )
    v_sun = 20 * u.km / u.s * SolarApex.galactic.cartesian
    U, V, W = v_sun.x, v_sun.y, v_sun.z
    print(U, V, W)
    v_bary = CartesianDifferential(U, V, W)
    return v_bary


class Observer:
    def __init__(self, location: EarthLocation):
        self.location = location

    def v_obs(
        self,
        lon: QuantityValue,
        lat: QuantityValue,
        frame: CoordFrameType,
        unit: Optional[UnitType] = None,
        time: Optional[float] = None,
    ) -> u.Quantity:
        time = pytime.time() if time is None else time
        unit = getattr(lon, "unit", None) if unit is None else unit
        unit = getattr(lat, "unit", "") if unit is None else unit

        time = Time(time, format="unix")
        target = SkyCoord(lon, lat, frame=frame, unit=unit)
        LSR_frame = LSR(v_bary=get_v_bary())
        v_obs = SkyCoord(self.location.get_gcrs(time)).transform_to(LSR_frame).velocity
        radial_velocity = v_obs.to_cartesian(target.cartesian).x
        return radial_velocity

    def lst(self, time: Optional[Union[float, Sequence[float]]] = None) -> Time:
        time = pytime.time() if time is None else time
        time = Time(time, format="unix", location=self.location)
        return time.sidereal_time("apparent")



from typing import Callable, Dict, List

import astropy.units as u

from ...utils import busy, sanity_check
from .ad_converter_base import ADConverter


class CPZ3177(ADConverter):
    Manufacturer = "Interface"
    Model = "CPZ3177"

    Identifier = "rsw_id"

    def __init__(self, **kwargs) -> None:
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.ave_num = self.Config.ave_num
        self.smpl_freq = self.Config.smpl_freq
        self.single_diff = self.Config.single_diff
        self.all_ch_num = self.Config.all_ch_num
        self.smpl_ch_req = self.Config.smpl_ch_req

        self.ad = pyinterface.open(3177, self.rsw_id)
        self.ad.stop_sampling()
        self.ad.initialize()
        self.ad.set_sampling_config(
            smpl_ch_req=self.smpl_ch_req,  # It must be a list
            smpl_num=1000,
            smpl_freq=self.smpl_freq,
            single_diff=self.single_diff,
            trig_mode="ETERNITY",
        )
        self.ad.start_sampling("ASYNC")

    def get_data(self, ch: int) -> List[float]:
        with busy(self, "busy"):
            offset = self.ad.get_status()["smpl_count"] - self.ave_num
            data = self.ad.read_sampling_buffer(self.ave_num, offset)
            data_li_2 = []
            for i in range(self.all_ch_num):
                data_li = []
                for k in range(self.ave_num):
                    data_li.append(data[k][i])
                data_li_2.append(data_li)

            ave_data_li = []
            for data in data_li_2:
                d = sum(data) / self.ave_num
                ave_data_li.append(d)
            return ave_data_li[ch - 1]

    @property
    def converter(self) -> Dict[str, Callable[[float], float]]:
        _ = [sanity_check(expr, "x") for expr in self.Config.converter.values()]
        return {k: eval(f"lambda x: {v}") for k, v in self.Config.converter.items()}

    def get_voltage(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        return self.converter["V"](self.get_data(ch)) * u.mV

    def get_current(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        return self.converter["I"](self.get_data(ch)) * u.microampere

    def get_power(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        return self.converter["P"](self.get_data(ch)) * u.mW

    def finalize(self) -> None:
        self.ad.stop_sampling()



import queue
import struct
import time
import traceback
from threading import Event, Thread
from typing import Dict, List, Tuple

import xfftspy

from ... import get_logger
from .spectrometer_base import Spectrometer


class XFFTS(Spectrometer):

    Manufacturer: str = "Radiometer Physics GmbH"
    Model: str = "XFFTS"

    Identifier = "host"

    def __init__(self, **kwargs) -> None:
        self.logger = get_logger(self.__class__.__name__)

        self.host = self.Config.host
        self.cmd_port = self.Config.cmd_port
        self.data_port = self.Config.data_port
        self.synctime_us = self.Config.synctime_us
        self.bw_mhz = {int(k): v for k, v in self.Config.bw_MHz.items()}
        self.data_input, self.setting_output = self.initialize()

        self.data_queue = queue.Queue(maxsize=2)
        self.thread = None
        self.event = None
        self.start()

        self.warn = False

    def start(self) -> None:
        if (self.thread is not None) or (self.event is not None):
            self.stop()
        self.thread = Thread(target=self._read_data, daemon=True)
        self.event = Event()
        self.thread.start()

    def _read_data(self) -> None:
        while (self.event is not None) and (not self.event.is_set()):
            if self.data_queue.full():
                if self.warn:
                    self.logger.warning(
                        "Dropping the data due to low readout frequency."
                    )
                    self.warn = False
                self.data_queue.get()

            try:
                data = self.data_input.receive_once()
                self.data_queue.put((time.time(), data["data"]))
            except struct.error:
                exc = traceback.format_exc()
                self.logger.warning(exc[slice(0, min(len(exc), 100))])

    def stop(self) -> None:
        if self.event is not None:
            self.event.set()
        if self.thread is not None:
            self.thread.join()
        self.event = self.thread = None
        self.warn = False

    def initialize(self) -> Tuple[xfftspy.data_consumer, xfftspy.udp_client]:
        """Get configured data input and setting output."""
        setting_output = xfftspy.udp_client(self.host, self.cmd_port, print=False)
        setting_output.stop()
        setting_output.set_synctime(self.synctime_us)  # synctime in us
        _sections = [int(i in self.bw_mhz) for i in range(1, max(self.bw_mhz) + 1)]
        setting_output.set_usedsections(_sections)
        for board_id, bw_mhz in self.bw_mhz.items():
            setting_output.set_board_bandwidth(board_id, bw_mhz)
        setting_output.configure()  # Apply settings
        setting_output.caladc()  # Calibrate ADCs
        setting_output.start()

        data_input = xfftspy.data_consumer(self.host, self.data_port)
        data_input.clear_buffer()
        return data_input, setting_output

    def get_spectra(self) -> Tuple[float, Dict[int, List[float]]]:
        self.warn = True
        return self.data_queue.get()

    def finalize(self) -> None:
        self.setting_output.stop()
        self.stop()
